
Cornmill = {};


function Cornmill.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(Overloading, specializations);
end

function Cornmill:preLoad(savegame)
    --self.getAllowPipeEffects = Utils.overwrittenFunction(self.getAllowPipeEffects, Cornmill.getAllowPipeEffects);
    --self.getOverloadingActive = Utils.overwrittenFunction(self.getOverloadingActive, Cornmill.getOverloadingActive);
end

function Cornmill:load(savegame)
    self.doOverload = Utils.overwrittenFunction(self.doOverload, Cornmill.doOverload);
    self.findValidTrailer = Utils.overwrittenFunction(self.findValidTrailer, Cornmill.findValidTrailer);
    self.findTrailerRaycastCallback = Utils.overwrittenFunction(self.findTrailerRaycastCallback, Cornmill.findTrailerRaycastCallback);
    Pipe.updateTick = Utils.overwrittenFunction(Pipe.updateTick, Cornmill.updateTickmill);
	self.ismill=true
	if self.isClient then
        self.pipeParticleSystems = {};
        self.cornmillconvert = {};
        local i=0;
        local j=0;
		
		
        while true do
            local key = string.format("vehicle.cornmill.cornmillconvert(%d)", i);
            if not hasXMLProperty(self.xmlFile, key) then
                break
            end
			
			local input= FillUtil.fillTypeNameToInt[getXMLString(self.xmlFile, key.."#input")]
			
			local outputfilltype= FillUtil.fillTypeNameToInt[getXMLString(self.xmlFile, key.."#output")]
			
			local ratefactor= Utils.getNoNil(getXMLFloat(self.xmlFile, key.."#rate"), 3) 
			
			
			if self.cornmillconvert[input] == nil then
				self.cornmillconvert[input] = {}
			end
			self.cornmillconvert[input].output= outputfilltype
			self.cornmillconvert[input].rate= ratefactor
			
			for _,fillUnit in pairs(self.fillUnits) do
                    for fillType,_ in pairs(fillUnit.fillTypes) do
						--print("filltype: "..tostring(fillType))
                        if fillType ~= nil then
							if self.cornmillconvert[fillType] == nil then
								self.cornmillconvert[fillType] = {}
							end
							if self.cornmillconvert[fillType].output== nil then
								self.cornmillconvert[fillType].output = fillType
								self.cornmillconvert[fillType].rate=1
							end
							
							if self.cornmillconvert[fillType].rate== nil then
								self.cornmillconvert[fillType].rate=1
							end
							
							
                        end
                    end
            end
						
            i=i+1;
		end	
			
		
		
        while true do
            local key = string.format("vehicle.pipeParticleSystems.emitterShape(%d)", j);
            if not hasXMLProperty(self.xmlFile, key) then
                break
            end

            local emitterShape = Utils.indexToObject(self.components, getXMLString(self.xmlFile, key.."#node"));
            local particleType = getXMLString(self.xmlFile, key.."#particleType")
            if emitterShape ~= nil then
                for _,fillUnit in pairs(self.fillUnits) do
                    for fillType,_ in pairs(fillUnit.fillTypes) do
						--print("filltype: "..tostring(fillType))
                        local particleSystem = MaterialUtil.getParticleSystem(self.cornmillconvert[fillType].output, particleType)
                        if particleSystem ~= nil then
                            if self.pipeParticleSystems[fillType] == nil then
                                self.pipeParticleSystems[fillType] = {}
                            end

                            local currentPS = ParticleUtil.copyParticleSystem(self.xmlFile, key, particleSystem, emitterShape)
                            local ps = currentPS
                            local normalSpeed,tangentSpeed = ParticleUtil.getParticleSystemAverageSpeed(ps);
                            ps.speed = math.sqrt(normalSpeed*normalSpeed + tangentSpeed*tangentSpeed);

                            table.insert(self.pipeParticleSystems[fillType], currentPS)
                        end
                    end
                end
            end

            j=j+1;
        end

    end

end

function Cornmill:update(dt)
end;

function Cornmill:draw()end;
function Cornmill:delete()end;
function Cornmill:mouseEvent(posX, posY, isDown, isUp, button)end;
function Cornmill:keyEvent(unicode, sym, modifier, isDown)end;

function Cornmill:updateTickmill(updateTick, dt)
    -- if a trailer was found, check (and adjust) control group of cylindered
    if self.isServer and self.hasPipe and self:getIsActive() then

        if self.pipeStateIsUnloading[self.pipeCurrentState] and self.trailerFound ~= 0 and self.pipeDisableManualMovement then

            if self.numControlGroups ~= nil and self.numControlGroups > 1 then

                local activeGroup = self.controlGroups[self.activeControlGroupIndex];
                if activeGroup.groupName == self.pipeControlGroupName then

                    local nextIndex = self:getNextValidControlGroupIndex();
                    if nextIndex ~= self.activeControlGroupIndex then
                        self:setActiveControlGroup(nextIndex);
                    end
                end
            end
        end

        self.pipeFoundTrailer = false;

        if not self:getOverloadingActive() then
            self:stopOverloadingDelay();
        end;
    end


    if self.isServer and (self:getUnitFillLevel(self.overloading.fillUnitIndex) > 0 or self:getUnitCapacity(self.overloading.fillUnitIndex) == 0) and self.pipeStateIsUnloading[self.pipeCurrentState] then

        if self.pipeEffects ~= nil then
            if self.pipeEffects[self.groundUnloadingPipeEffect] ~= nil then
                local controlPointY = self.pipeEffects[self.groundUnloadingPipeEffect].controlPointY;
                local distance = self.pipeEffects[self.groundUnloadingPipeEffect].distance;

                if controlPointY ~= nil and distance ~= nil and controlPointY ~= 0 then
                    local mCos = math.cos(controlPointY);
                    local mSin = math.sin(controlPointY);

                    local y = Utils.dotProduct(0, 0, distance, 0.0, mCos, -mSin);
                    local z = Utils.dotProduct(0, 0, distance, 0.0, mSin, mCos);

                    if self.pipeStateIsUnloading[self.pipeTargetState] then
                        local x, y, z = localToWorld(self.pipeEffects[self.groundUnloadingPipeEffect].node, 0, y, z);
                        self.groundUnloadingPosition = {x, y, z};
                    end;
                end;
            end;
        end;

        -- do raycast
        self.trailerFound = 0;
        self.shovelTargetFound = nil;
        self.trailerFoundDistance = 0;
        if self.pipeRaycastNode ~= nil then
            local doRaycast = self.pipeStateIsUnloading[self.pipeTargetState];

            if doRaycast then
                local x,y,z = getWorldTranslation(self.pipeRaycastNode);
                local dx,dy,dz = localDirectionToWorld(self.pipeRaycastNode, 0,-1,0);
                raycastAll(x, y, z, dx,dy,dz, "findTrailerRaycastCallback", self.pipeRaycastDistance, self);
            end

            if self.trailerFoundDistance == 0 then
                self.trailerFound = 0;
            end;

            if self.targetTrailer ~= nil then
                if self.targetTrailer.coverAnimation ~= nil then
                    if not self.targetTrailer.isCoverOpen then
                        self.trailerFound = 0;
                    end;
                end;
            end;
        end

        if self.trailerFound ~= 0 then
            if self.dischargeToGround then
                self:setDischargeToGround(false);
            end
        else
            if self.dischargeToGround then
                if not self:getCanTipToGround() or (self.overloading.stopOverloadIfEmpty and (self:getUnitLastValidFillType(self.overloading.fillUnitIndex) == FillUtil.FILLTYPE_UNKNOWN or self:getUnitFillLevel(self.overloading.fillUnitIndex) == 0)) then
                    self:setDischargeToGround(false);
                end
            end
        end
		
        if self.dischargeToGround then
            if self:getUnitFillLevel(self.overloading.fillUnitIndex) > 0 and not self:getOverloadingDelayIsActive() then
                self:startOverloadingDelay();
            end;

            local node = self.pipeRaycastNode;
            if self.pipeDischargeInfoIndex ~= nil then
                node = self.fillVolumeDischargeInfos[self.pipeDischargeInfoIndex].nodes[1].node;
            end;

            local speed = 10;
            local x,y,z = Utils.getIntersectionOfLinearMovementAndTerrain(node, speed);
            local x0,y0,z0 = getWorldTranslation(node);
            self.pipeUnloadingDistance = Utils.vector3Length(x0-x, y0-y, z0-z);

            if self.groundUnloadingPosition[1] ~= 0 then
                x,y,z = unpack(self.groundUnloadingPosition);
            end;

            if self:getAllowDelayedUnloading() then
                local fillType = self:getUnitLastValidFillType(self.overloading.fillUnitIndex);
				local millfillType = fillType
				local millrate = 1
				
				if self.cornmillconvert == nil then
				else
					if self.cornmillconvert[fillType] == nil then
					else
							millfillType=self.cornmillconvert[fillType].output;
							millrate=self.cornmillconvert[fillType].rate;
							if millfillType == nil then
								millfillType = fillType
							end
							if millrate == nil then
								millrate = 1
							end
					end
				end
				
                local fillLevel = self:getUnitFillLevel(self.overloading.fillUnitIndex);

                local deltaFill = fillLevel;
                if self.lastLostFillLevel ~= nil then
                    deltaFill = math.max(fillLevel, self.lastLostFillLevel);
                end

                if self:getUnitCapacity(self.overloading.fillUnitIndex) > 0 then
                    deltaFill = math.min(deltaFill, self.overloading.capacity*dt/1000.0);
                end

                self.remainingFillLevelToGround = math.min(self.remainingFillLevelToGround + deltaFill, math.max(self.overloading.capacity*dt/1000.0, TipUtil.getMinValidLiterValue(millfillType)));

                if self.remainingFillLevelToGround > 0 then
                    local infos = self.fillVolumeDischargeInfos[self.overloading.dischargeInfoIndex];
                    local dx,dy,dz;
                    if infos ~= nil then
                        dx,dy,dz = localDirectionToWorld(infos.nodes[1].node, infos.nodes[1].width,0,0);
                    else
                        dx,dy,dz = localDirectionToWorld(self.components[1].node, 0,0,0.25);
                    end

                    local sx, sy, sz = x+dx, y+dy, z+dz;
                    local ex, ey, ez = x-dx, y-dy, z-dz;

                    if self.limitDischargeToPipeHeight then
                        if infos ~= nil then
                            local x,y,z = localToWorld(infos.nodes[1].node, infos.nodes[1].width,0,0);
                            sy = y;
                            local x,y,z = localToWorld(infos.nodes[1].node, -infos.nodes[1].width,0,0);
                            ey = y;
                        else
                            local x,y,z = getWorldTranslation(self.pipeRaycastNode);
                            sy = y;
                            ey = y;
                        end
                    else
                    end

                    local dropped, lineOffset = TipUtil.tipToGroundAroundLine(self, self.remainingFillLevelToGround, millfillType, sx,sy,sz, ex,ey,ez, 0.1, nil, self.tipToGroundLineOffset, self.limitDischargeToPipeHeight, self.tipOcclusionAreas);

                    self.tipToGroundLineOffset = lineOffset;

                    self.remainingFillLevelToGround = self.remainingFillLevelToGround - dropped;
					if millrate == 1 then
                    self:setUnitFillLevel(self.overloading.fillUnitIndex, fillLevel - (dropped), fillType, false, self.fillVolumeUnloadInfos[self.overloading.unloadInfoIndex]);
					else					
                    self:setUnitFillLevel(self.overloading.fillUnitIndex, fillLevel - (dropped/millrate), fillType, false, self.fillVolumeUnloadInfos[self.overloading.unloadInfoIndex]);
					end
					
                    self.overloading.didOverload = dropped > 0;
                end
            end;

            if not self:getCanTipToGround() or (self.overloading.stopOverloadIfEmpty and (self:getUnitLastValidFillType(self.overloading.fillUnitIndex) == FillUtil.FILLTYPE_UNKNOWN or self:getUnitFillLevel(self.overloading.fillUnitIndex) == 0)) then
                self:setDischargeToGround(false);
            end
        elseif not self.dischargeToGround and self.overloading.isActive then
            self.pipeUnloadingDistance = self.trailerFoundDistance;
        end
    else
        self.trailerFound = 0;
    end

    if self.isServer then
        if self.pipeUnloadingDistance ~= self.sentPipeUnloadingDistance or self.sentPipeFoundTrailer ~= self.pipeFoundTrailer then
            self:raiseDirtyFlags(self.pipeDirtyFlag);

            self.sentPipeUnloadingDistance = self.pipeUnloadingDistance;
            self.sentPipeFoundTrailer = self.pipeFoundTrailer;
        end
    end

    if self.isClient then
        self.overloading.delay.time = self.pipeBaseDelay
        if self.delayEffect ~= nil then
            self.overloading.delay.time = self.delayEffect.startDelay + (1-self.delayEffect.offset)*self.delayEffect.fadeInTime
        end
        if self:getIsActiveForInput(self) then
            if self.pipeOffsets ~= nil then
                for _, offset in pairs(self.pipeOffsets) do
                    local movingTool = offset.movingTool
                    local state = Cylindered.getMovingToolState(self, movingTool);

                    for _, effect in pairs(offset.effects) do
                        local effectState = state
                        if effect.inverted then
                            effectState = 1 - effectState
                        end
                        effect.effect:setOffset(Utils.lerp(effect.minValue, effect.maxValue, effectState))
                        if effect.adjustDelay then
                            local delay = Utils.lerp(effect.minDelay, effect.maxDelay, effectState)
                            effect.effect:setDelays(delay, delay)
                        end
                    end
                end
            end
        end

        if self:getAllowDelayEffects() then
            self.pipeParticleDeactivateTime = g_currentMission.time + self.pipeUnloadEffectStopTime;

            local currentPipeParticleSystems = self.pipeParticleSystems[self:getUnitLastValidFillType(self.overloading.fillUnitIndex)];
            if currentPipeParticleSystems ~= self.currentPipeParticleSystems then
                if self.currentPipeParticleSystems ~= nil then
                    for _, ps in pairs(self.currentPipeParticleSystems) do
                        ParticleUtil.setEmittingState(ps, false)
                    end
                end
                self.currentPipeParticleSystems = currentPipeParticleSystems
                if self.currentPipeParticleSystems ~= nil then
                    for _, ps in pairs(self.currentPipeParticleSystems) do
                        ParticleUtil.setEmittingState(ps, true)
                    end
                end
            end

            if self.currentPipeParticleSystems ~= nil then
                for _, currentPS in pairs(self.currentPipeParticleSystems) do
                    local ps = currentPS
                    if not ps.forceFullLifespan then
                        local factor = 1+((8-self.pipeUnloadingDistance-self.pipeParticleSystemExtraDistance)/15);
                        local lifespan = ps.originalLifespan * (self.pipeUnloadingDistance+self.pipeParticleSystemExtraDistance) * factor;
                        ParticleUtil.setParticleLifespan(ps, lifespan)
                    end
                end
            end

            if self.pipeEffects ~= nil then
			
				local millfillTypeeffect = self:getUnitLastValidFillType(self.overloading.fillUnitIndex)
				
				local filltypecheck=self:getUnitLastValidFillType(self.overloading.fillUnitIndex)
				if self.cornmillconvert == nil then
				else	
					if self.cornmillconvert[filltypecheck] == nil then
					else
						millfillTypeeffect=self.cornmillconvert[filltypecheck].output;
						if millfillTypeeffect == nil then
							millfillTypeeffect = self:getUnitLastValidFillType(self.overloading.fillUnitIndex)
						end
					end	
				end
                EffectManager:setFillType(self.pipeEffects, millfillTypeeffect)
                --EffectManager:setFillType(self.pipeEffects, self:getUnitLastValidFillType(self.overloading.fillUnitIndex))
                EffectManager:startEffects(self.pipeEffects)
            end
        end

        if self.pipeEffects ~= nil then
            for _, effect in pairs(self.pipeEffects) do
                if effect.setUnloadingDistance ~= nil then
                    effect:setUnloadingDistance(self.pipeUnloadingDistance+self.pipeParticleSystemExtraDistance, g_currentMission.terrainRootNode)
                end
            end;
        end

        if self.pipeParticleDeactivateTime >= 0 and self.pipeParticleDeactivateTime <= g_currentMission.time then
            self.pipeParticleDeactivateTime = -1;
            if self.currentPipeParticleSystems ~= nil then
                for _, ps in pairs(self.currentPipeParticleSystems) do
                    ParticleUtil.setEmittingState(ps, false)
                end
                self.currentPipeParticleSystems = nil;
            end

            if self.pipeEffects ~= nil then
                EffectManager:stopEffects(self.pipeEffects);
            end
        end
    
	end
end



function Cornmill:doOverload(superFunc, targetTrailer, fillType, deltaFill)

	local millfillType = fillType
	local millrate = 1

    if self.pipeRaycastNode == nil then
        return superFunc(self, targetTrailer, millfillType, deltaFill);
    end

	if self.cornmillconvert == nil then
	else
		if self.cornmillconvert[fillType] ==nil then
		else
			millfillType=self.cornmillconvert[fillType].output;
			millrate=self.cornmillconvert[fillType].rate;
				if millfillType == nil then
					millfillType = fillType
				end
				if millrate == nil then
					millrate = 1
				end
		end	
	end	
	
    local trailer = g_currentMission.nodeToVehicle[self.trailerFound];
    if targetTrailer == trailer and targetTrailer:getAllowFillFromAir() and deltaFill > 0 then
        targetTrailer:resetFillLevelIfNeeded(millfillType);

        local trailerFillLevel = targetTrailer:getFillLevel(millfillType);
        local trailerCapacity = targetTrailer:getCapacity(millfillType);

        deltaFill = math.min(deltaFill, trailerCapacity - trailerFillLevel);
        if deltaFill > 0 then
            targetTrailer:setFillLevel(trailerFillLevel+(deltaFill*millrate), millfillType, false, self.fillVolumeDischargeInfos[self.overloading.dischargeInfoIndex]);
            local newFillLevel = targetTrailer:getFillLevel(millfillType);
            local diff = (newFillLevel - trailerFillLevel)/millrate;
            self:setUnitFillLevel(self.overloading.fillUnitIndex, self:getUnitFillLevel(self.overloading.fillUnitIndex)-diff, fillType, false, self.fillVolumeUnloadInfos[self.overloading.unloadInfoIndex]);
            return true;
        else
            return false;
        end
    else
        return false;
    end
end


function Cornmill:findValidTrailer(superFunc, trailers)



--millfillType=FillUtil.FILLTYPE_CCMRAW;
--local fillType=FillUtil.FILLTYPE_CCMRAW;
    if self.pipeRaycastNode == nil then
        return superFunc(self, trailers);
    end

    local nearestTrailer;
	
	local fillTypecheck = self:getUnitLastValidFillType(self.overloading.fillUnitIndex);
	local fillType=fillTypecheck

	if self.cornmillconvert == nil then
	else
		if self.cornmillconvert[fillTypecheck] ==nil then
		else
			fillType=self.cornmillconvert[fillTypecheck].output;
			if fillType == nil then
				fillType = self:getUnitLastValidFillType(self.overloading.fillUnitIndex)
			end
		end
	end

    --local fillType = millfillType;
   -- local fillType = self:getUnitLastValidFillType(self.overloading.fillUnitIndex);

    -- check if a trailer is attached
    for _,implement in pairs(self.attachedImplements) do
        if implement.object ~= nil then
            if implement.object.exactFillRootNode ~= nil then
                for trailer,value in pairs(trailers) do
                    if value > 0 then
                        if entityExists(trailer.components[1].node) then
                            if trailer == implement.object and trailer:allowFillType(fillType) and trailer:getFillLevel(fillType) < trailer:getCapacity(fillType) then
                                nearestTrailer = trailer;
                                break;
                            end
                        end
                    end
                end
            end
        end
        if nearestTrailer ~= nil then
            break;
        end
    end

    -- find closest trailer to raycast node in triggers
    if nearestTrailer == nil then
	        local nearestDistance = math.huge;
	        local x0,y0,z0 = getWorldTranslation(self.pipeRaycastNode);
	
	        for trailer,value in pairs(trailers) do
	            if not entityExists(trailer.components[1].node) then
	                value = 0;
	            end
	            if value > 0 then
	                if trailer:allowFillType(fillType) and trailer.exactFillRootNode ~= nil then
	                    if trailer:getFillLevel(fillType) < trailer:getCapacity(fillType) then
	                        local x1,y1,z1 = getWorldTranslation(trailer.exactFillRootNode);
	                        local dist = Utils.vector3LengthSq(x0-x1, y0-y1, z0-z1);
	                        if dist < nearestDistance then
	                            nearestTrailer = trailer;
	                            nearestDistance = dist;
	                        end
	                    end
	                end
	            end
	        end
	    end
	
	    return nearestTrailer;
	end
	
	function Cornmill:findTrailerRaycastCallback(superFunc, transformId, x, y, z, distance)
	    if self.pipeRaycastNode == nil then
			return superFunc(self, transformId, x, y, z, distance);
		end
	    local vehicle = g_currentMission.nodeToVehicle[transformId];
	    if vehicle ~= nil and vehicle ~= self then
	        if self.trailerFound ~= 0 and self.trailerFoundDistance ~= 0 then
	            return false;
	        end;
	
	        self.pipeFoundTrailer = true;
	
	        if vehicle.exactFillRootNode == transformId then

				local fillTypecheck = self:getUnitLastValidFillType(self.overloading.fillUnitIndex);
				local fillType=fillTypecheck
			
				if self.cornmillconvert == nil then
				else	
					if self.cornmillconvert[fillTypecheck] == nil then
					else
						fillType=self.cornmillconvert[fillTypecheck].output;
						if fillType == nil then
							fillType = self:getUnitLastValidFillType(self.overloading.fillUnitIndex)
						end
					end	
				end
			 -- local fillType = self:getUnitLastValidFillType(self.overloading.fillUnitIndex);
	            if vehicle:allowFillType(fillType) and vehicle.exactFillRootNode ~= nil then
	                if vehicle:getFillLevel(fillType) < vehicle:getCapacity(fillType) then
	                    self.trailerFound = transformId;
	                end
	            end
	        else
	            self.trailerFoundDistance = distance+self.pipeParticleSystemExtraDistanceOnTrailer;
	        end
	    else
	        local object = g_currentMission:getNodeObject(transformId);
	        if object ~= nil and object.parent ~= nil and object.parent.triggerId ~= nil then
	            if object.parent:getAllowFillTypeFromTool(self:getUnitLastValidFillType(self.overloading.fillUnitIndex), TipTrigger.TOOL_TYPE_PIPE) then
	                self.shovelTargetFound = object;
	                self.trailerFoundDistance = distance+self.pipeParticleSystemExtraDistanceOnTrailer;
	            end;
	        end;
	    end
		

	
	    return true;
	end
	